#!/usr/bin/env python3
# vim:set ts=4 sts=4 sw=4 et:
"""Simple reconstruction algorithms.

This module defines a base class, ``Reconstruction``.
"""
from __future__ import division, absolute_import, print_function

import numpy as np
import pandas as pd     # noqa
from scipy.linalg import eigvals
from scipy.stats import (
    chisquare, kurtosis, skew, entropy, mode, kurtosistest, normaltest,
    skewtest, tmean, tvar, tstd,
)

from statsmodels.robust.scale import mad

from km3pipe import Module
from km3pipe.dataclasses import KM3Array, KM3DataFrame     # noqa
from km3pipe.tools import azimuth, zenith, dist, unit_vector
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103


try:
    from scipy.stats import iqr
except ImportError:
    log.warning("Please update your scipy installation in order to use the "
                "`iqr` function. It will now return `np.nan`.")
    iqr = lambda x: np.nan


def bimod(skew, kurt, n, is_fisher=True):
    """Test a distribution for bimodality."""
    if not is_fisher:
        kurt = kurt - 3
    return np.divide(
        np.square(skew) + 1,
        kurt + np.divide(3 * np.square(n - 1), (n - 2) * (n - 3))
    )


def uniform_chi2(x):
    """Test a distribution for uniformity."""
    chi2, _ = chisquare(x, f_exp=None)
    return chi2


def idr(x, perc=10):
    up, lo = np.percentile(x, [100-perc, perc])
    return up - lo


def tensor_of_inertia(self, pos_x, pos_y, pos_z, weight=1):
    """Tensor of Inertia.

    Adapted from Thomas Heid' EventID (ROOT implementation).
    """
    toi = np.zeros((3, 3), dtype=float)
    toi[0][0] += np.square(pos_y) * np.square(pos_z)
    toi[0][1] += (-1) * pos_x * pos_y
    toi[0][2] += (-1) * pos_x * pos_z
    toi[1][0] += (-1) * pos_x * pos_y
    toi[1][1] += np.squar(pos_x) + np.square(pos_z)
    toi[1][2] += (-1) * pos_y * pos_z
    toi[2][0] += (-1) * pos_x * pos_z
    toi[2][1] += (-1) * pos_z * pos_y
    toi[2][2] += np.square(pos_x) + np.square(pos_y)
    toi *= weight
    return eigvals(toi)


class Reconstruction(Module):
    """Reconstruction base class.

    Parameters
    ----------
    hit_sel: str, default='Hits'
        Blob key of the hits to run the fit on.
    key_out: str, default='GenericReco'
        Key to write into.
    """

    def __init__(self, **kwargs):
        super(Reconstruction, self).__init__(**kwargs)
        self.hit_sel = self.get('hit_sel') or 'Hits'
        self.key_out = self.get('key_out') or 'GenericReco'

    def process(self, blob):
        hits = blob[self.hit_sel].serialise(to='pandas')
        reco = self.fit(hits)
        if reco is not None:
            blob[self.key_out] = KM3DataFrame.deserialise(
                reco, h5loc='/reco', fmt='dict')
        return blob

    def fit(self, hits):
        return


class SvdFit(Reconstruction):
    """Primitive Linefit using Singular Value Decomposition.

    Parameters
    ----------
    n_doms_min: int, default=8
        Minimum number of Doms with hits.  Events with fewer will not
        be reconstructed.
    hit_sel: str, default='Hits'
        Blob key of the hits to run the fit on.
    key_out: str, default='SvdFit'
        Key to write into.

    Returns
    -------
    out: None (fit failed), or KM3DataFrame
        The km3df keys are: ['pos_x,y,z', 'dir_x,y,z', 'phi', 'theta'].
    """
    def __init__(self, **kwargs):
        super(SvdFit, self).__init__(**kwargs)
        self.hit_sel = self.get('hit_sel') or 'Hits'
        self.key_out = self.get('key_out') or 'SvdFit'
        self.n_doms_min = self.get('n_doms_min') or 8
        self.n_doms_min = int(self.n_doms_min)

    def fit(self, hits):
        dus = np.unique(hits['dom_id'])
        n_dus = len(dus)

        if n_dus < 8:
            return

        pos = hits[['pos_x', 'pos_y', 'pos_z']]
        center = pos.mean(axis=0)

        _, _, v = np.linalg.svd(pos - center)

        reco_dir = np.array(v[0])
        reco_pos = center
        len = dist(pos.iloc[-1], pos.iloc[0])
        dur = hits.time.iloc[-1] - hits.time.iloc[0]
        velo = len/dur

        out = {
            'pos_x': reco_pos['pos_x'],
            'pos_y': reco_pos['pos_y'],
            'pos_z': reco_pos['pos_z'],
            'dir_x': reco_dir[0],
            'dir_y': reco_dir[1],
            'dir_z': reco_dir[2],
            'phi': azimuth(reco_dir),
            'theta': zenith(reco_dir),
            'speed': velo,
        }
        return out


class LineFit(Reconstruction):
    """Fit a linear track, with direction and velocity.

    Parameters
    ----------
    hit_sel: str, default='Hits'
        Blob key of the hits to run the fit on.
    key_out: str, default='LineFit'
        Key to write into.

    Returns
    -------
    out: None (fit failed), or KM3DataFrame
        The km3df keys are: ['pos_x,y,z', 'dir_x,y,z', 'phi', 'theta', '
        speed'].
    """
    def __init__(self, **kwargs):
        super(LineFit, self).__init__(**kwargs)
        self.hit_sel = self.get('hit_sel') or 'Hits'
        self.key_out = self.get('key_out') or 'LineFit'

    def fit(self, hits):
        pos = hits[['pos_x', 'pos_y', 'pos_z']]

        reco_dir = unit_vector(pos.iloc[-1] - pos.iloc[0])

        out = {
            'dir_x': reco_dir[0],
            'dir_y': reco_dir[1],
            'dir_z': reco_dir[2],
            'phi': azimuth(reco_dir),
            'theta': zenith(reco_dir),
            'speed': velo,
        }
        return out


class Trawler(Reconstruction):
    """Compute summary statistics on hits.
    """
    def __init__(self, **kwargs):
        super(Trawler, self).__init__(**kwargs)
        self.hit_sel = self.get('hit_sel') or 'Hits'
        self.key_out = self.get('key_out') or 'Trawler'

    def fit(self, hits):
        out = {}

        n_hits = len(hits)
        out['n_hits'] = n_hits
        r2 = np.sqrt(
            np.square(hits.pos_x) +
            np.square(hits.pos_y)
        )
        r3 = np.sqrt(
            np.square(hits.pos_x) +
            np.square(hits.pos_y) +
            np.square(hits.pos_z)
        )
        z = hits.pos_z
        out['tot_sum'] = np.sum(hits.tot)
        out['tot_max'] = np.max(hits.tot)
        out['tot_max_over_nhits'] = out['tot_max'] / n_hits
        out['tot_max_rel'] = np.max(hits.tot) / np.sum(hits.tot)
        out['tot_sum_over_nhits'] = out['tot_sum'] / n_hits
        out['tot_sum_over_max'] = out['tot_sum'] / out['tot_max']
        out['z_totmax'] = hits.pos_z[np.argmax(hits.tot)]

        out.update(self.stats(z, 'z'))
        out.update(self.stats(r2, 'r2'))
        out.update(self.stats(r3, 'r3'))
        out.update(self.stats(hits.time, 'time'))

        # first pulses / truncated: variance etc

        # time windows (n_win = 3)
        for deno in ('tot_sum', 'n_hits', 'tot_max'):
            for meas in ('var', 'cog', 'idr', 'iqr'):
                for nume in ('r2', 'r3', 'z'):
                    feat = nume + '_' + meas + '_over_' + deno
                    out[feat] = out[nume + '_' + meas] / out[deno]
        return out

    @classmethod
    def stats(cls, x, prefix='', wgt=None):
        if wgt is None:
            wgt = np.ones_like(x)
        n_hits = len(x)
        o = {}
        o['med'] = np.median(x)
        o['mean'] = np.average(x)
        o['var'] = np.var(x)
        o['skew'] = skew(x)
        o['kurt'] = kurtosis(x, fisher=True)
        o['mad'] = mad(x)
        o['sum'] = np.sum(x)
        o['iqr'] = iqr(x)
        o['idr'] = idr(x)
        o['uni_chi2'] = uniform_chi2(x)
        o['bimod'] = bimod(o['skew'], o['kurt'], n_hits)
        o['entropy'] = entropy(x)
        o['cog'] = np.average(x, weights=wgt)
        o = {prefix + '_' + key: val for key, val in o.items()}     # noqa
        return o



#!/usr/bin/env python
# coding=utf-8
# Filename: setup.py
"""
KM3Pipe setup script.

"""
from setuptools import setup, Extension
import sys

if sys.version_info[0] >= 3:
    import builtins
else:
    import __builtin__ as builtins
try:
    from Cython.Distutils import build_ext
    from Cython.Build import cythonize
    # from Cython.Compiler.Options import directive_defaults
    import numpy
except ImportError:
    raise SystemExit("\nCython and Numpy are required to compile KM3Pipe.\n"
                     "You can install it easily via pip:\n\n"
                     "    > pip install cython numpy")


# This hack is "stolen" from numpy and allows to detect the setup procedure
# to avoid loading modules which are not compiled yet. Ugly but robust.
builtins.__KM3PIPE_SETUP__ = True

# Needed for line_profiler - disable for production code
directives = {
    'linetrace': True,
    'profile': True,
    'binding': True,
}
CYTHON_TRACE = '1'

from km3pipe import version  # noqa


tools = Extension('km3pipe.tools', sources=['km3pipe/tools.pyx'],
                  extra_compile_args=['-O3', '-march=native', '-w'],
                  include_dirs=[numpy.get_include()],
                  define_macros=[('CYTHON_TRACE', CYTHON_TRACE)])

core = Extension('km3pipe.core', sources=['km3pipe/core.pyx'],
                 extra_compile_args=['-O3', '-march=native', '-w'],
                 include_dirs=[numpy.get_include()],
                 define_macros=[('CYTHON_TRACE', CYTHON_TRACE)])

dataclasses = Extension('km3pipe.dataclasses',
                        sources=['km3pipe/dataclasses.pyx'],
                        extra_compile_args=['-O3', '-march=native', '-w'],
                        include_dirs=[numpy.get_include()],
                        define_macros=[('CYTHON_TRACE', CYTHON_TRACE)])

setup(name='km3pipe',
      version=version,
      url='http://github.com/tamasgal/km3pipe/',
      description='An analysis framework for KM3NeT',
      author='Tamas Gal and Moritz Lotze',
      author_email='tgal@km3net.de',
      packages=['km3pipe', 'km3pipe.testing', 'km3pipe.io', 'km3pipe.utils',
                'km3modules', 'pipeinspector'],
      ext_modules=cythonize([tools, core, dataclasses],
                            compiler_directives=directives),
      cmdclass={'build_ext': build_ext},
      include_package_data=True,
      platforms='any',
      install_requires=[
          'cython',
          'docopt',
          'matplotlib>=2.0.0',
          'mock',
          'numpy>=1.12',
          'pandas',
          'pytz',
          'scipy',
          'six',
          'tables',
          'h5py',
          'tornado',
          'urwid',
          'websocket-client',
      ],
      extras_require={
          'docs': ['sphinx>=1.5.1', 'sphinx-gallery', 'numpydoc',
                   'scikit-learn', 'statsmodels>=0.8', 'seaborn',
                   'pillow', 'ipython', 'sphinxcontrib-napoleon',
                   'astropy', 'h5py', 'patsy'],
          'jppy': ['jppy'],
          'testing': ['pytest', 'mock'],
      },
      entry_points={
          'console_scripts': [
              'km3pipe=km3pipe.cmd:main',
              'km3srv=km3pipe.srv:main',
              'tohdf5=km3pipe.utils.tohdf5:main',
              'hdf2root=km3pipe.utils.hdf2root:main',
              'pipeinspector=pipeinspector.app:main',
              'h5tree=km3pipe.utils.h5tree:main',
              'rtree=km3pipe.utils.rtree:main',
              'h5info=km3pipe.utils.h5info:main',
              'h5concat=km3pipe.utils.h5concat:main',
              'meantots=km3pipe.utils.meantots:main',
              'pushover=km3pipe.utils.pushover:main',
              'ztplot=km3pipe.utils.ztplot:main',
              'k40calib=km3pipe.utils.k40calib:main',
          ],
      },
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Programming Language :: Python',
      ],
      )

__author__ = 'Tamas Gal'



# coding=utf-8
# Filename: test_db.py
# pylint: disable=C0111,E1003,R0904,C0103,R0201,C0102
from __future__ import division, absolute_import, print_function

from km3pipe.testing import TestCase, MagicMock

from km3pipe.db import DBManager, DOMContainer
from km3pipe.logger import logging

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

DET_ID = 'det_id1'
JSON_DOMS = [{'DOMId': 1, 'Floor': 10, 'CLBUPI': '100', 'DetOID': DET_ID},
             {'DOMId': 2, 'Floor': 20, 'CLBUPI': '200', 'DetOID': DET_ID},
             {'DOMId': 3, 'Floor': 30, 'CLBUPI': '300', 'DetOID': DET_ID},
             {'DOMId': 4, 'Floor': 40, 'CLBUPI': '400', 'DetOID': 'det_id2'}]

log = logging.getLogger('db')


class TestDBManager(TestCase):

    def test_login_called_on_init_when_credentials_are_provided(self):
        user = 'user'
        pwd = 'god'

        DBManager.login = MagicMock()
        db = DBManager(username=user, password=pwd, temporary=True)
        self.assertEqual(1, DBManager.login.call_count)
        self.assertTupleEqual((user, pwd), DBManager.login.call_args[0])

    def test_login(self):
        original_login = DBManager.login  # save for later
        user = 'a'
        pwd = 'b'

        # mock login to be able to create an instance without an actual login
        DBManager.login = MagicMock()
        db = DBManager(username='foo', password='bar')  # make dummy call
        DBManager.login = original_login  # restore function

        db._make_request = MagicMock()
        db.login(username='a', password='b')
        call_args = db._make_request.call_args[0]
        self.assertEqual(db._login_url, call_args[0])
        self.assertDictEqual({'usr': 'a', 'pwd': 'b'}, call_args[1])


class TestDOMContainer(TestCase):
    def test_init(self):
        DOMContainer(None)

    def setUp(self):
        self.dc = DOMContainer(JSON_DOMS)

    def test_ids_returns_dom_ids(self):
        self.assertListEqual([1, 2, 3], self.dc.ids(DET_ID))

    def test_json_list_lookup(self):
        lookup = self.dc._json_list_lookup('DOMId', 1, 'Floor', DET_ID)
        self.assertEqual(10, lookup)

    def test_clbupi2floor(self):
        self.assertEqual(10, self.dc.clbupi2floor('100', DET_ID))
        self.assertEqual(20, self.dc.clbupi2floor('200', DET_ID))
        self.assertEqual(30, self.dc.clbupi2floor('300', DET_ID))

    def test_clbupi2domid(self):
        self.assertEqual(1, self.dc.clbupi2domid('100', DET_ID))
        self.assertEqual(2, self.dc.clbupi2domid('200', DET_ID))
        self.assertEqual(3, self.dc.clbupi2domid('300', DET_ID))



# coding=utf-8
# Filename: __init__.py
"""
The extemporary KM3NeT analysis framework.

"""
from __future__ import division, absolute_import, print_function


try:
    __KM3PIPE_SETUP__
except NameError:
    __KM3PIPE_SETUP__ = False

from .__version__ import version, version_info  # noqa

if not __KM3PIPE_SETUP__:
    from .core import (Pipeline, Module, Pump, Blob, Run,  # noqa
                              Geometry, AanetGeometry)
    from . import io  # noqa
    from . import utils  # noqa
    from . import srv  # noqa
    from .srv import srv_event  # noqa
    from .io import GenericPump, read_hdf5  # noqa


__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__version__ = version
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"



# coding=utf-8
# Filename: rba.py
"""
RainbowAlga online display.

Usage:
    rba -t TOKEN -n EVENT_ID [-u URL] FILE
    rba (-h | --help)
    rba --version

Options:
    FILE       Input file.
    -h --help  Show this screen.

"""
from __future__ import division, absolute_import, print_function

import tables

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def rba():
    pass


def main():
    from docopt import docopt
    arguments = docopt(__doc__)

    rba(arguments['FILE'])



# coding=utf-8
# Filename: tohdf5.py
"""
Send a push message to a device using Pushover.net API.

Usage:
    pushover MESSAGE...
    pushover (-h | --help)
    pushover --version

Options:
    MESSAGE     The message to send.
    -h --help   Show this screen.

"""
from __future__ import division, absolute_import, print_function

from km3pipe.config import Config
from km3pipe import version

try:
    import http.client as httplib
    from urllib.parse import urlencode
except ImportError:
    import httplib
    from urllib import urlencode

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"



def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)

    config = Config()

    token = config.get("Pushover", "token")
    user_key = config.get("Pushover", "user_key")
    if token is None or user_key is None:
        print("Please define your 'token' and 'user_key' in the "
              "'Pushover' section of your ~/.km3net configuration.")
        exit(1)

    conn = httplib.HTTPSConnection("api.pushover.net:443")
    conn.request("POST", "/1/messages.json",
      urlencode({
        "token": token,
        "user": user_key,
        "message": ' '.join(args["MESSAGE"]),
      }), { "Content-type": "application/x-www-form-urlencoded" })

    conn.getresponse()



# coding=utf-8
# Filename: hardware.py
# pylint: disable=locally-disabled
"""
Classes representing KM3NeT hardware.

"""
from __future__ import division, absolute_import, print_function

from collections import OrderedDict, defaultdict
import os
import sys
from six import iteritems
from six.moves import xrange

import numpy as np

from .tools import unpack_nfirst, split, com  # , ignored
from .dataclasses import Point, Direction
from .db import DBManager

from .logger import logging

if sys.version_info[0] > 2:
    from io import StringIO
else:
    from StringIO import StringIO


log = logging.getLogger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class Detector(object):
    """A KM3NeT detector.

    Parameters
    ----------
    filename: str, optional
        Name of .detx file with detector definition.
    det_id: int, optional
        .detx ID of detector (when retrieving from database).
    t0set: optional
        t0set (when retrieving from database).
    calibration: optional
        calibration (when retrieving from database).
    """
    def __init__(self, filename=None,
                 det_id=None,
                 t0set=None,
                 calibration=None):
        self._det_file = None
        self.det_id = None
        self.n_doms = None
        self.lines = set()
        self.n_pmts_per_dom = None
        self.doms = OrderedDict()
        self.pmts = []
        self.version = None
        self.valid_from = None
        self.valid_until = None
        self.utm_info = None
        self._dom_ids = []
        self._dom_positions = OrderedDict()
        self._pmts_by_omkey = OrderedDict()
        self._pmts_by_id = OrderedDict()
        self._pmts_by_dom_id = defaultdict(list)
        self._pmt_angles = []

        if filename:
            self._init_from_file(filename)

        if det_id is not None:
            print("Retrieving DETX with detector ID {0} from the database..."
                  .format(det_id))
            db = DBManager()
            detx = db.detx(det_id, t0set=t0set, calibration=calibration)
            self._det_file = StringIO(detx)
            self._parse_header()
            self._parse_doms()
            if self.n_doms < 1:
                log.error("No data found for detector ID {0}.".format(det_id))

    def _init_from_file(self, filename):
        """Create detector from detx file."""
        file_ext = os.path.splitext(filename)[1][1:]
        if not file_ext == 'detx':
            raise NotImplementedError('Only the detx format is supported.')
        self._open_file(filename)
        self._parse_header()
        self._parse_doms()
        self._det_file.close()

    def _open_file(self, filename):
        """Create the file handler"""
        self._det_file = open(filename, 'r')

    def _parse_header(self):
        """Extract information from the header of the detector file"""
        self._det_file.seek(0, 0)
        first_line = self._det_file.readline()
        try:
            self.det_id, self.n_doms = split(first_line, int)
            self.version = 'v1'
        except ValueError:
            det_id, self.version = first_line.split()
            self.det_id = int(det_id)
            validity = self._det_file.readline()
            self.valid_from, self.valid_until = split(validity, float)
            self.utm_info = self._det_file.readline()
            n_doms = self._det_file.readline()
            self.n_doms = int(n_doms)

    # pylint: disable=C0103
    def _parse_doms(self):
        """Extract dom information from detector file"""
        self._det_file.seek(0, 0)
        self._det_file.readline()
        lines = self._det_file.readlines()
        try:
            while True:
                line = lines.pop(0)
                if not line:
                    continue
                try:
                    dom_id, line_id, floor_id, n_pmts = split(line, int)
                except ValueError:
                    continue
                self.lines.add(line_id)
                self.n_pmts_per_dom = n_pmts
                for i in range(n_pmts):
                    raw_pmt_info = lines.pop(0)
                    pmt_info = raw_pmt_info.split()
                    pmt_id, x, y, z, rest = unpack_nfirst(pmt_info, 4)
                    dx, dy, dz, t0, rest = unpack_nfirst(rest, 4)
                    if rest:
                        log.warn("Unexpected PMT values: '{0}'".format(rest))
                    pmt_id = int(pmt_id)
                    pmt_pos = [float(n) for n in (x, y, z)]
                    pmt_dir = [float(n) for n in (dx, dy, dz)]
                    t0 = float(t0)
                    if floor_id < 0:
                        _, new_floor_id, _ = self._pmtid2omkey_old(pmt_id)
                        log.error("Floor ID is negative for PMT {0}.\n"
                                  "Guessing correct id: {1}"
                                  .format(pmt_id, new_floor_id))
                        floor_id = new_floor_id
                    # TODO: following line is here bc of the bad MC floor IDs
                    #      put it outside the for loop in future
                    self.doms[dom_id] = (line_id, floor_id, n_pmts)
                    omkey = (line_id, floor_id, i)
                    pmt = PMT(pmt_id, pmt_pos, pmt_dir, t0, i, omkey)
                    self.pmts.append(pmt)
                    self._pmts_by_omkey[(line_id, floor_id, i)] = pmt
                    self._pmts_by_id[pmt_id] = pmt
                    self._pmts_by_dom_id[dom_id].append(pmt)
        except IndexError:
            pass

    @property
    def dom_ids(self):
        if not self._dom_ids:
            self._dom_ids = self.doms.keys()
        return self._dom_ids

    @property
    def dom_positions(self):
        """The positions of the DOMs, calculated as COM from PMTs."""
        if not self._dom_positions:
            for dom_id in self.dom_ids:
                pmt_positions = [p.pos for p in self._pmts_by_dom_id[dom_id]]
                self._dom_positions[dom_id] = com(pmt_positions)
        return self._dom_positions

    def translate_detector(self, vector):
        vector = np.array(vector, dtype=float)
        for pmt in self.pmts:
            pmt.pos = pmt.pos + vector

    @property
    def pmt_angles(self):
        """A list of PMT directions sorted by PMT channel"""
        if not self._pmt_angles:
            pmts = self.pmts[:self.n_pmts_per_dom]
            self._pmt_angles = [pmt.dir for pmt in pmts]
        return self._pmt_angles

    @property
    def ascii(self):
        """The ascii representation of the detector"""
        if self.version == 'v1':
            header = "{det.det_id} {det.n_doms}".format(det=self)
        else:
            header = "{det.det_id} {det.version}".format(det=self)
            header += "\n{0} {1}".format(self.valid_from, self.valid_until)
            header += "\n" + self.utm_info
            header += str(self.n_doms)

        doms = ""
        for dom_id, (line, floor, n_pmts) in iteritems(self.doms):
            doms += "{0} {1} {2} {3}\n".format(dom_id, line, floor, n_pmts)
            for i in xrange(n_pmts):
                pmt = self._pmts_by_omkey[(line, floor, i)]
                doms += "{0} {1} {2} {3} {4} {5} {6} {7}\n".format(
                        pmt.id, pmt.pos[0], pmt.pos[1], pmt.pos[2],
                        pmt.dir[0], pmt.dir[1], pmt.dir[2],
                        pmt.t0
                        )
        return header + "\n" + doms

    def write(self, filename):
        """Save detx file."""
        with open(filename, 'w') as f:
            f.write(self.ascii)
        print("Detector file saved as '{0}'".format(filename))

    def pmt_with_id(self, pmt_id):
        """Get PMT with global pmt_id"""
        try:
            return self._pmts_by_id[pmt_id]
        except KeyError:
            raise KeyError("No PMT found for ID: {0}".format(pmt_id))

    def get_pmt(self, dom_id, channel_id):
        """Return PMT with DOM ID and DAQ channel ID"""
        line, floor, _ = self.doms[dom_id]
        pmt = self._pmts_by_omkey[(line, floor, channel_id)]
        return pmt

    def pmtid2omkey(self, pmt_id):
        return self._pmts_by_id[int(pmt_id)].omkey

    def _pmtid2omkey_old(self, pmt_id,
                         first_pmt_id=1, oms_per_line=18, pmts_per_om=31):
        """Convert (consecutive) raw PMT IDs to Multi-OMKeys."""
        pmts_line = oms_per_line * pmts_per_om
        line = ((pmt_id - first_pmt_id) // pmts_line) + 1
        om = oms_per_line - (pmt_id - first_pmt_id) % pmts_line // pmts_per_om
        pmt = (pmt_id - first_pmt_id) % pmts_per_om
        return int(line), int(om), int(pmt)

    def domid2floor(self, dom_id):
        _, floor, _ = self.doms[dom_id]
        return floor

    @property
    def n_lines(self):
        return len(self.lines)

    def __str__(self):
        return "Detector id: '{0}', n_doms: {1}, n_lines: {2}".format(
            self.det_id, self.n_doms, self.n_lines)

    def __repr__(self):
        return self.__str__()


class PMT(object):
    """Represents a photomultiplier.

    Parameters
    ----------
    id: int
    pos: 3-float-tuple (x, y, z)
    dir: 3-float-tuple (x, y, z)
    t0: int
    channel_id: int
    omkey: int
    """
    def __init__(self, id, pos, dir, t0, channel_id, omkey):
        self.id = id
        self.pos = Point(pos)
        self.dir = Direction(dir)
        self.t0 = t0
        self.channel_id = channel_id
        self.omkey = omkey

    def __str__(self):
        return "PMT id:{0}, pos: {1}, dir: dir{2}, t0: {3}, DAQ channel: {4}"\
               .format(self.id, self.pos, self.dir, self.t0, self.channel_id)



#!/usr/bin/env python3
# vim:set ts=4 sts=4 sw=4 et:
"""MC Helpers.
"""
from __future__ import division, absolute_import, print_function

from km3pipe import Module
from km3pipe.tools import zenith, azimuth, pdg2name, name2pdg


NEUTRINOS = {'nu_e', 'anu_e', 'nu_mu', 'anu_mu', }      # noqa


class McTruth(Module):
    """Extract MC info of 1st MC track.

    Parameters
    ----------
    most_energetic_primary: bool, default=True
    """
    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        self.most_energetic_primary = bool(self.get('most_energetic_primary')) or True

    @classmethod
    def t2f(cls, row):
        return pdg2name(row['type'])

    @classmethod
    def is_nu(cls, flavor):
        return flavor in NEUTRINOS

    def process(self, blob):
        mc = blob['McTracks'].conv_to('pandas')
        if self.most_energetic_primary:
            mc.sort_values('energy', ascending=False, inplace=True)
        mc = mc.head(1)
        mc['zenith'] = zenith(mc[['dir_x', 'dir_y', 'dir_z']])
        mc['azimuth'] = azimuth(mc[['dir_x', 'dir_y', 'dir_z']])
        flavor = mc.apply(self.t2f, axis=1)
        mc['is_neutrino'] = flavor.apply(self.is_nu)
        blob['McTruth'] = mc
        return blob



# coding=utf-8
# Filename: pandas.py
# pylint: disable=C0103,R0903
# vim:set ts=4 sts=4 sw=4 et:
"""
Pandas Helpers.
"""
from __future__ import division, absolute_import, print_function

import os.path
from six import string_types

import numpy as np
import pandas as pd
import tables as tb

from km3pipe.logger import logging
from km3pipe.tools import insert_prefix_to_dtype

log = logging.getLogger(__name__)  # pylint: disable=C0103

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


class H5Chain(object):
    """Read/write multiple HDF5 files as ``pandas.DataFrame``.

    It is impliend that all files share the same group/tables
    structure and naming.

    Parameters
    ----------
    filenames: list(str), or dict(fname -> h5file)

    Examples
    --------

    >>> filenames = ['numu_cc.h5', 'anue_nc.h5']
    >>> c = H5Chain(filenames)
    >>> X = c['/reco/gandalf']

    A context manager is also available:

    >>> with H5Chain(filenames) as h5:
    >>>     reco = h5['/reco']

    """
    def __init__(self, filenames):
        self.h5files = {}
        for fn in filenames:
            h5 = pd.HDFStore(fn, 'r')
            self.h5files[fn] = h5

    def close(self):
        for h5 in self.h5files.values():
            h5.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __enter__(self):
        return self

    def __getitem__(self, key):
        dfs = []
        for fname, h5 in self.h5files.items():
            df = h5[key]
            dfs.append(df)
        dfs = pd.concat(dfs, axis=0, ignore_index=True)
        return dfs

    def __call__(self, key):
        """
        Parameters
        ----------
        key: str
            H5 path of the object to retrieve, e.g. '/reco/gandalf'.
        """
        return self[key]


def map2df(map):
    return pd.DataFrame.from_records(map, index=np.ones(1, dtype=int))


def _read_table(tab, max_id=None, ignore_events=False):
    if ignore_events:
        return tab[:max_id]
    else:
        return tab.read_where('event_id <= %d' % max_id)


def read_group(group, max_id=None, **kwargs):
    # Store through groupname, insert tablename into dtype
    df = []
    for tab in group._f_iter_nodes(classname='Table'):
        tabname = tab.name
        if max_id is None:
            arr = tab.read(**kwargs)
        else:
            arr = _read_table(tab, max_id)
        arr = insert_prefix_to_dtype(arr, tabname)
        arr = pd.DataFrame.from_records(arr)
        df.append(arr)
    df = pd.concat(df, axis=1)
    return df


def merge_event_ids(df):
    cols = list(df.columns)
    ids = list(c for c in cols if 'event_id' in c)
    if not ids:
        return df
    # non_id = list(c for c in cols if c not in ids)
    event_ids = df[ids[0]]
    df.drop(ids, axis=1, inplace=True)
    df['event_id'] = event_ids
    return df


def df_to_h5(df, h5file, where):
    """Write pandas dataframes with proper columns.

    Example:
        >>> df = pd.DataFrame(...)
        >>> df_to_h5(df, 'foo.h5', '/some/loc/my_df')
    """
    write_table(df.to_records(index=False), h5file, where)


def write_table(array, h5file, where):
    """Write a structured numpy array into a H5 table.
    """
    own_h5 = False
    if isinstance(h5file, string_types):
        own_h5 = True
        h5file = tb.open_file(h5file, 'a')
    filt = tb.Filters(complevel=5, shuffle=True, fletcher32=True)
    loc, tabname = os.path.split(where)
    if loc == '':
        loc = '/'
    h5file.create_table(loc, tabname, obj=array, createparents=True,
                        filters=filt)
    if own_h5:
        h5file.close()


def first_mc_tracks(mc_tracks, mupage=False):
    mc_tracks = pd.DataFrame(mc_tracks)
    if mupage:
        mc_tracks = mc_tracks[mc_tracks.type != 0]
        mc_tracks = mc_tracks[mc_tracks.id == 1]
    return mc_tracks.drop_duplicates(subset='event_id')



# coding=utf-8
# Filename: plot.py
# pylint: disable=locally-disabled
"""
A collection of plotting functions and modules.

"""
from __future__ import division, absolute_import, print_function

from datetime import datetime
import os

import matplotlib
# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import pylab
import numpy as np

import km3pipe as kp
import km3pipe.style


def plot_dom_parameters(data, detector, filename, label, title,
                        vmin=0.0, vmax=10.0,
                        cmap='RdYlGn_r', under='deepskyblue', over='deeppink',
                        underfactor=1.0, overfactor=1.0,
                        missing='lightgray'):
    """Creates a plot in the classical monitoring.km3net.de style.

    Parameters
    ----------
    data: dict((du, floor) -> value)
    detector: km3pipe.hardware.Detector() instance
    filename: filename or filepath
    label: str
    title: str
    underfactor: a scale factor for the points used for underflow values
    overfactor: a scale factor for the points used for overflow values

    """
    x, y, _ = zip(*detector.doms.values())
    fig, ax = plt.subplots(figsize=(10, 6))
    cmap = plt.get_cmap(cmap)
    cmap.set_over(over, 1.0)
    cmap.set_under(under, 1.0)

    m_size = 100
    scatter_args = {
        'edgecolors': 'None',
        'vmin': vmin,
        'vmax': vmax,
    }
    sc_inactive = ax.scatter(x, y, c=missing, label='missing',
                             s=m_size*0.9,
                             **scatter_args)

    xa, ya = map(np.array, zip(*data.keys()))
    zs = np.array(list(data.values()))
    in_range_idx = np.logical_and(zs>=vmin, zs<=vmax)
    sc = ax.scatter(xa[in_range_idx], ya[in_range_idx],
                    c=zs[in_range_idx], cmap=cmap, s=m_size,
                    **scatter_args)
    under_idx = zs < vmin
    sc_under = ax.scatter(xa[under_idx], ya[under_idx],
                         c=under, label='< {0}'.format(vmin),
                         s=m_size*underfactor,
                         **scatter_args)
    over_idx = zs > vmax
    sc_over = ax.scatter(xa[over_idx], ya[over_idx],
                         c=over, label='> {0}'.format(vmax),
                         s=m_size*overfactor,
                         **scatter_args)

    cb = plt.colorbar(sc)
    cb.set_label(label)

    ax.set_title("{0}\n{1} UTC".format(title, datetime.utcnow().strftime("%c")))
    ax.set_xlabel("DU")
    ax.set_ylabel("DOM")
    ax.set_ylim(-2)
    ax.set_yticks(range(1, 18+1))
    major_locator = pylab.MaxNLocator(integer=True)
    sc_inactive.axes.xaxis.set_major_locator(major_locator)

    ax.legend(bbox_to_anchor=(0., -.16 , 1., .102), loc=1,
               ncol=2, mode="expand", borderaxespad=0.)

    fig.tight_layout()

    plt.savefig(filename, dpi=120, bbox_inches="tight")
    plt.close('all')
